from datetime import datetime, timedelta
import base64
import json
import os
import random
import requests
import time
from time import sleep

import logging
logging.basicConfig(
    format='%(asctime)s %(levelname)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    level=logging.DEBUG
)

class SimpleBeatSaverAPI:
    BASE_URL = "https://api.beatsaver.com"

    def __init__(self, cache_expiry_days=1):
        self.session = requests.Session()
        self.cache_expiry_days = cache_expiry_days
        self.CACHE_DIR = self._determine_cache_dir()
        if not os.path.exists(self.CACHE_DIR):
            os.makedirs(self.CACHE_DIR)

    def _determine_cache_dir(self):
        home_cache = os.path.expanduser("~/.cache")
        saberlist_cache = os.path.join(home_cache, "saberlist")
        beatsaver_cache = os.path.join(saberlist_cache, "beatsaver")
        
        if os.path.exists(home_cache):
            if not os.path.exists(saberlist_cache):
                try:
                    os.makedirs(saberlist_cache)
                    logging.info(f"Created cache directory: {saberlist_cache}")
                except OSError as e:
                    logging.warning(f"Failed to create {saberlist_cache}: {e}")
                    return os.path.join(os.getcwd(), ".cache")
            return beatsaver_cache
        else:
            logging.info("~/.cache doesn't exist, using local .cache directory")
            return os.path.join(os.getcwd(), ".cache")

    def _get_cache_filename(self, player_id):
        return os.path.join(self.CACHE_DIR, f"player_{player_id}_scores.json")

    def _is_cache_valid(self, cache_file):
        if not os.path.exists(cache_file):
            return False
        file_modified_time = datetime.fromtimestamp(os.path.getmtime(cache_file))
        return datetime.now() - file_modified_time < timedelta(days=self.cache_expiry_days)

    def clear_cache(self, player_id=None):
        if player_id:
            cache_file = self._get_cache_filename(player_id)
            if os.path.exists(cache_file):
                os.remove(cache_file)
                logging.debug(f"Cleared cache for player {player_id}")
        else:
            for file in os.listdir(self.CACHE_DIR):
                os.remove(os.path.join(self.CACHE_DIR, file))
            logging.debug("Cleared all cache")

    def get_cache_dir(self):
        return self.CACHE_DIR


    def get_curated_songs(self, use_cache=True) -> list[dict]:
        """
        Retrieve curated songs from BeatSaver.

        Note: BeatSaver's sort order behavior is not documented, so we cannot reliably
        cache results per page or determine if new songs have been added. Therefore,
        we must fetch all pages each time to ensure we have the complete, up-to-date list.

        :param use_cache: Whether to use cached data if available (default: True)
        :return: List of dictionaries containing song hash, characteristic, and difficulty
        """
        cache_file = os.path.join(self.CACHE_DIR, f"curated_songs.json")

        # Never expire the curated songs cache because it results in a many calls to the API
        if use_cache and os.path.exists(cache_file):
            logging.debug(f"Using cached data for curated songs")
            with open(cache_file, 'r') as f:
                return json.load(f)

        processed_songs = []
        page = 0

        while True:
            url = f"{self.BASE_URL}/search/text/{page}"
            params = {
                "sortOrder": "Curated",
                "curated": "true"
            }

            try:
                response = self.session.get(url, params=params)
                response.raise_for_status()
                data = response.json()
                
                # Process the response to extract relevant information
                for song in data.get('docs', []):
                    for version in song.get('versions', []):
                        song_info = {
                            'hash': version['hash'],
                            'key': song['id'],
                            'songName': song['metadata']['songName'],
                        }
                        processed_songs.append(song_info)
                
                # Check if we've reached the last page
                if page >= data['info']['pages'] - 1:
                    break
                    
                page += 1
                sleep(1)

            except requests.exceptions.RequestException as e:
                logging.error(f"Error fetching curated songs: {e}")
                return []

        # Cache the results
        with open(cache_file, 'w') as f:
            json.dump(processed_songs, f)
        
        return processed_songs

    def get_followed_mappers(self, user_id: int = 243016, use_cache=True) -> list[dict]:
        """
        Retrieve list of mappers followed by a specific user.
        
        :param user_id: BeatSaver user ID
        :param use_cache: Whether to use cached data if available (default: True)
        :return: List of dictionaries containing mapper information
        """
        cache_file = os.path.join(self.CACHE_DIR, f"followed_mappers_{user_id}.json")
        
        if use_cache and self._is_cache_valid(cache_file):
            logging.debug(f"Using cached data for followed mappers of user {user_id}")
            with open(cache_file, 'r') as f:
                return json.load(f)

        url = f"{self.BASE_URL}/users/followedBy/{user_id}/0"
        
        try:
            response = self.session.get(url)
            response.raise_for_status()
            mappers = response.json()
            
            # Cache the results
            with open(cache_file, 'w') as f:
                json.dump(mappers, f)
            
            return mappers

        except requests.exceptions.RequestException as e:
            logging.error(f"Error fetching followed mappers: {e}")
            return []

    def get_mapper_maps(self, mapper_id: int, use_cache=True) -> list[dict]:
        """
        Retrieve all maps created by a specific mapper.
        
        :param mapper_id: BeatSaver mapper ID
        :param use_cache: Whether to use cached data if available (default: True)
        :return: List of dictionaries containing map information
        """
        cache_file = os.path.join(self.CACHE_DIR, f"mapper_{mapper_id}_maps.json")

        if use_cache and self._is_cache_valid(cache_file):
            logging.debug(f"Using cached data for mapper {mapper_id}")
            with open(cache_file, 'r') as f:
                return json.load(f)

        processed_songs = []
        page = 0

        while True:
            url = f"{self.BASE_URL}/search/text/{page}"
            params = {
                'collaborator': str(mapper_id),
                'automapper': 'true',
                'sortOrder': 'Latest'
            }

            try:
                response = self.session.get(url, params=params)
                response.raise_for_status()
                data = response.json()
                
                # Process the response to extract relevant information
                for song in data.get('docs', []):
                    for version in song.get('versions', []):
                        song_info = {
                            'hash': version['hash'],
                            'key': song['id'],
                            'songName': song['metadata']['songName'],
                        }
                        processed_songs.append(song_info)
                
                # Page numbering starts at 0
                if page >= data['info']['pages'] - 1:
                    break
                    
                page += 1
                sleep(1)

            except requests.exceptions.RequestException as e:
                logging.error(f"Error fetching mapper maps: {e}")
                return []

        # Cache the results
        with open(cache_file, 'w') as f:
            json.dump(processed_songs, f)
        
        return processed_songs
